#!/usr/bin/env python3

"""
Command-line entry point
"""

# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111,w1202,r1705


# DON'T import logger globally here
# we need to be able to mess inside the logger module
# before it gets loaded from another way

import asyncio

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

from pkg_resources import resource_string

from asynciojobs import Scheduler, Job

import logging
from asyncssh.logging import set_log_level as asyncssh_set_log_level

from rhubarbe.config import Config
from rhubarbe.imagesrepo import ImagesRepo
from rhubarbe.selector import (Selector,
                               add_selector_arguments, selected_selector)
from rhubarbe.action import Action
from rhubarbe.display import Display
from rhubarbe.display_curses import DisplayCurses
from rhubarbe.node import Node
from rhubarbe.imageloader import ImageLoader
from rhubarbe.imagesaver import ImageSaver
from rhubarbe.monitor.nodes import MonitorNodes
from rhubarbe.monitor.loop import MonitorLoop
from rhubarbe.monitor.phones import MonitorPhones
from rhubarbe.monitor.leases import MonitorLeases
from rhubarbe.monitor.accountsmanager import AccountsManager
from rhubarbe.ssh import SshProxy
from rhubarbe.leases import Leases
from rhubarbe.inventory import Inventory
from rhubarbe.inventoryphones import InventoryPhones


# a supported command comes with a driver function
# in this module, that takes a list of args
# and returns 0 for success and s/t else otherwise
# specifically, command
# rhubarbe load -i fedora 12
# would result in a call
# load ( "-i", "fedora", "12" )


####################
RESERVATION_REQUIRED = "This function requires a valid reservation - "\
                       "or to be root or a privileged user"


####################
# exposed to the outside world (typically r2lab's nightly)

def check_reservation(leases, *,                        # pylint: disable=w0621
                      root_allowed=True, verbose=False, login=None):
    """
    return a bool indicating if a login currently has a lease

    if login is None, the current login is used instead

    when True, root_allowed means that the root user is always granted access

    verbose can be
    None  : does not write anything
    False : write a message if lease is not there
    True : always write a message
    """

    async def check_leases():
        # login = None means use my login
        # we don't use booked_now_by_me because of the verbose message
        # we use another variable to avoid using nonlocal
        actual_login = login or leases.login
        if verbose:
            print(f"Checking current reservation for {actual_login} : ", end="")
        is_fine = await leases.booked_now_by(login=actual_login,
                                             root_allowed=root_allowed)
        if is_fine:
            if verbose:
                print("OK")
        else:
            if verbose is None:
                pass
            elif verbose:
                print(f"WARNING: Access currently denied to {actual_login}")
            else:
                print("access denied")
        return is_fine
    return asyncio.get_event_loop().run_until_complete(check_leases())


def no_reservation(leases):                             # pylint: disable=w0621
    """
    returns True if nobody currently has a lease
    """
    async def check_leases():
        return not await leases.booked_now_by_anyone()
    return asyncio.get_event_loop().run_until_complete(check_leases())


####################
# NOTE: when adding a new command, please update setup.py as well
supported_subcommands = []                              # pylint: disable=c0103


def subcommand(driver):
    supported_subcommands.append(driver.__name__)
    return driver

####################


@subcommand
def nodes(*argv):
    usage = """
    Just display the hostnames of selected nodes
    """
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    add_selector_arguments(parser)
    args = parser.parse_args(argv)
    selector = selected_selector(args)
    print(" ".join(selector.node_names()))

####################


def cmc_verb(verb, resa_policy, *argv):
    """
    resa_policy can be either
    (*) 'enforce': refuse to send the message if the lease is not there
    (*) 'warn': issue a warning when the lease is not there
    (*) 'none' - or anything else really: does not check the leases
    """
    usage = f"""
    Send verb '{verb}' to the CMC interface of selected nodes"""
    if resa_policy == 'enforce':
        usage += f"\n    {RESERVATION_REQUIRED}"
    config = Config()
    default_timeout = config.value('nodes', 'cmc_default_timeout')

    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-t", "--timeout", action='store',
                        default=default_timeout, type=float,
                        help="Specify global timeout for the whole process")
    add_selector_arguments(parser)
    args = parser.parse_args(argv)

    message_bus = asyncio.Queue()
    leases = Leases(message_bus)                        # pylint: disable=w0621

    if resa_policy in ('warn', 'enforce'):
        reserved = check_reservation(leases, verbose=False)
        if not reserved:
            if resa_policy == 'enforce':
                return 1

    selector = selected_selector(args)
    action = Action(verb, selector)

    return 0 if action.run(message_bus, args.timeout) else 1

#####


@subcommand
def status(*argv):
    return cmc_verb('status', 'warn', *argv)


@subcommand
def on(*argv):                                          # pylint: disable=c0103
    return cmc_verb('on', 'enforce', *argv)


@subcommand
def off(*argv):
    return cmc_verb('off', 'enforce', *argv)


@subcommand
def reset(*argv):
    return cmc_verb('reset', 'enforce', *argv)


@subcommand
def info(*argv):
    return cmc_verb('info', 'warn', *argv)


@subcommand
def usrpstatus(*argv):
    return cmc_verb('usrpstatus', 'warn', *argv)


@subcommand
def usrpon(*argv):
    return cmc_verb('usrpon', 'enforce', *argv)


@subcommand
def usrpoff(*argv):
    return cmc_verb('usrpoff', 'enforce', *argv)


#####
# xxx should be asynchronous
@subcommand
def bye(*argv):
    """
    An alternative implementation of the previous 'all-off' utility
    Switch the lights off when you leave
    """

    usage = """
    Turn off whole testbed
    """
    config = Config()
    default_timeout = config.value('nodes', 'cmc_default_timeout')
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-t", "--timeout", action='store',
                        default=default_timeout, type=float,
                        help="Specify timeout for each phase")
    add_selector_arguments(parser)
    args = parser.parse_args(argv)

    selector = selected_selector(args, defaults_to_all=True)
    if selector.is_empty():
        selector.use_all_scope()

    bus = asyncio.Queue()
    Action('usrpoff', selector).run(bus, args.timeout)

    # keep it simple for now
    import time
    time.sleep(1)
    Action('off', selector).run(bus, args.timeout)

    # even simpler
    import os
    phones_inventory = InventoryPhones()
    for phone in phones_inventory.all_phones():
        command = (f"ssh -i {phone['gw_key']} "
                   f"{phone['gw_user']}@{phone['gw_host']} phone-off")
        print(command)
        os.system(command)

####################


@subcommand
def load(*argv):
    usage = f"""
    Load an image on selected nodes in parallel
    {RESERVATION_REQUIRED}
    """
    config = Config()
    config.check_binaries()
    imagesrepo = ImagesRepo()

    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-i", "--image", action='store',
                        default=imagesrepo.default(),
                        help="Specify image to load")
    parser.add_argument("-t", "--timeout", action='store',
                        default=config.value('nodes',
                                                 'load_default_timeout'),
                        type=float,
                        help="Specify global timeout for the whole process")
    parser.add_argument("-b", "--bandwidth", action='store',
                        default=config.value('networking', 'bandwidth'),
                        type=int,
                        help="Set bandwidth in Mibps for frisbee uploading")
    parser.add_argument("-c", "--curses", action='store_true', default=False,
                        help="Use curses to provide term-based animation")
    # this is more for debugging
    parser.add_argument("-n", "--no-reset", dest='reset',
                        action='store_false', default=True,
                        help="""use this with nodes that are already
                        running a frisbee image. They won't get reset,
                        neither before or after the frisbee session""")
    add_selector_arguments(parser)
    args = parser.parse_args(argv)

    message_bus = asyncio.Queue()

    selector = selected_selector(args)
    if selector.is_empty():
        parser.print_help()
        return 1
    nodes = [Node(cmc_name, message_bus)                # pylint: disable=w0621
             for cmc_name in selector.cmc_names()]

    # send feedback
    message_bus.put_nowait({'selected_nodes': selector})
    from rhubarbe.logger import logger
    logger.info(f"timeout is {args.timeout}s")
    logger.info(f"bandwidth is {args.bandwidth} Mibps")

    actual_image = imagesrepo.locate_image(args.image, look_in_global=True)
    if not actual_image:
        print(f"Image file {args.image} not found - emergency exit")
        exit(1)

    # send feedback
    message_bus.put_nowait({'loading_image': actual_image})
    display_class = Display if not args.curses else DisplayCurses
    display = display_class(nodes, message_bus)
    loader = ImageLoader(nodes, image=actual_image, bandwidth=args.bandwidth,
                         message_bus=message_bus, display=display)
    return loader.main(reset=args.reset, timeout=args.timeout)

####################


@subcommand
def save(*argv):
    usage = f"""
    Save an image from a node
    Mandatory radical needs to be provided with --output
      This info, together with nodename and date, is stored
      on resulting image in /etc/rhubarbe-image
    {RESERVATION_REQUIRED}
    """

    config = Config()
    config.check_binaries()

    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-o", "--output", action='store', dest='radical',
                        default=None, required=True,
                        help="Mandatory radical to name resulting image")
    parser.add_argument("-t", "--timeout", action='store',
                        default=config.value('nodes',
                                                 'save_default_timeout'),
                        type=float,
                        help="Specify global timeout for the whole process")
    parser.add_argument("-c", "--comment", dest='comment', default=None,
                        help="one-liner comment to insert in "
                        "/etc/rhubarbe-image")
    parser.add_argument("-n", "--no-reset", dest='reset',
                        action='store_false', default=True,
                        help="""use this with a node that is already
                        running a frisbee image. It won't get reset,
                        neither before or after the frisbee session""")
    parser.add_argument("node")
    args = parser.parse_args(argv)

    message_bus = asyncio.Queue()

    selector = Selector()
    selector.add_range(args.node)
    # in case there was one argument but it was not found in inventory
    if len(selector) != 1:
        parser.print_help()
    cmc_name = next(selector.cmc_names())
    node = Node(cmc_name, message_bus)
    nodename = node.control_hostname()

    imagesrepo = ImagesRepo()
    actual_image = imagesrepo.where_to_save(nodename, args.radical)
    message_bus.put_nowait({'info': f"Saving image {actual_image}"})
    # curses has no interest here since we focus on one node
    display_class = Display
    display = display_class([node], message_bus)
    saver = ImageSaver(node, image=actual_image, radical=args.radical,
                       message_bus=message_bus, display=display,
                       comment=args.comment)
    return saver.main(reset=args.reset, timeout=args.timeout)

####################


@subcommand
def wait(*argv):                                        # pylint: disable=r0914
    usage = """
    Wait for selected nodes to be reachable by ssh
    Returns 0 if all nodes indeed are reachable
    """
    # suppress info log messages from asyncssh
    asyncssh_set_log_level(logging.WARNING)

    config = Config()
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-c", "--curses", action='store_true', default=False,
                        help="Use curses to provide term-based animation")
    parser.add_argument("-t", "--timeout", action='store',
                        default=config.value('nodes',
                                                 'wait_default_timeout'),
                        type=float,
                        help="Specify global timeout for the whole process")
    parser.add_argument("-b", "--backoff", action='store',
                        default=config.value('networking', 'ssh_backoff'),
                        type=float,
                        help="Specify backoff average between "
                        "attempts to ssh connect")
    # really dont' write anything
    parser.add_argument("-s", "--silent", action='store_true', default=False)
    parser.add_argument("-v", "--verbose", action='store_true', default=False)

    add_selector_arguments(parser)
    args = parser.parse_args(argv)

    # --curses implies --verbose otherwise nothing shows up
    if args.curses:
        args.verbose = True

    selector = selected_selector(args)
    message_bus = asyncio.Queue()

    if args.verbose:
        message_bus.put_nowait({'selected_nodes': selector})
    from rhubarbe.logger import logger
    logger.info(f"wait: backoff is {args.backoff} "
                f"and global timeout is {args.timeout}")

    nodes = [Node(cmc_name, message_bus)                # pylint: disable=w0621
             for cmc_name in selector.cmc_names()]
    sshs = [SshProxy(node, verbose=args.verbose) for node in nodes]
    jobs = [Job(ssh.wait_for(args.backoff), critical=True) for ssh in sshs]

    display_class = Display if not args.curses else DisplayCurses
    display = display_class(nodes, message_bus)

    # have the display class run forever until the other ones are done
    scheduler = Scheduler(Job(display.run(), forever=True, critical=True),
                          *jobs,
                          timeout=args.timeout,
                          critical=False)
    try:
        orchestration = scheduler.run()
        if orchestration:
            return 0
        else:
            if args.verbose:
                scheduler.debrief()
            return 1
    except KeyboardInterrupt:
        print("rhubarbe-wait : keyboard interrupt - exiting")
        # xxx
        return 1
    finally:
        display.epilogue()
        if not args.silent:
            for ssh in sshs:
                print(f"{ssh.node}:ssh {'OK' if ssh.status else 'KO'}")

####################


@subcommand
def images(*argv):
    usage = """
    Display available images
    """
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-l", "--labeled-only", dest="labeled",
                        action='store_true', default=False,
                        help="""if specified, only images that have at least
                        one symlink to them are shown""")
    parser.add_argument("-p", "--public-only",
                        action="store_true", default=False,
                        help="displays only publicly visible images")
    parser.add_argument("-s", "--size", dest='sort_size',
                        action='store_true', default=None,
                        help="sort by size (default is by name)")
    parser.add_argument("-d", "--date", dest='sort_date',
                        action='store_true', default=None,
                        help="sort by date")
    parser.add_argument("-r", "--reverse",
                        action='store_true', default=False,
                        help="reverse sort")
    parser.add_argument("-n", "--narrow",
                        action='store_true', default=False,
                        help="""default is to show full paths, with this option
                        only radicals are displayed""")
    parser.add_argument("focus", nargs="*", type=str,
                        help="if provided, only images that contain "
                        "one of these strings are displayed")
    args = parser.parse_args(argv)
    imagesrepo = ImagesRepo()
    if args.sort_size is not None:
        args.sort_by = 'size'
    elif args.sort_date is not None:
        args.sort_by = 'date'
    else:
        args.sort_by = 'name'

    # if focus is an empty list, then everything is shown
    return imagesrepo.images(
        args.focus, args.sort_by, args.reverse,
        args.labeled, args.public_only, args.narrow)

####################


@subcommand
def resolve(*argv):
    usage = """for each input, find out and display
    what file exactly would be used if used with load
    and possible siblings if verbose mode is selected
    """

    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
#    parser.add_argument("-r", "--reverse", action='store_true', default=False,
#                        help="reverse sort")
    parser.add_argument("-v", "--verbose", action='store_true', default=False,
                        help="show all files (i.e. with all symlinks)")
    parser.add_argument("focus", type=str,
                        help="the image radical name to resolve")
    args = parser.parse_args(argv)
    imagesrepo = ImagesRepo()
    # if focus is an empty list, then everything is shown
    return imagesrepo.resolve(args.focus, args.verbose)

####################


@subcommand
def share(*argv):
    usage = """
    Install privately-stored images into the global images repo
    Destination name is derived from the radical provided at save-time
      i.e. trailing part after saving__<date>__

    With the -a option, it is possible to make the new image
    **also** visible under that alias name.

    If your account is enabled in /etc/sudoers.d/rhubarbe-share,
    the command will actually perform the mv operation
    Requires to be run through 'sudo rhubarbe-share'
    """
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-a", "--alias-name", dest='alias',
                        action='store', default=None,
                        help="also create a symlink of that name")
    # default=None so that imagesrepo.share can compute a default
    parser.add_argument("-n", "--dry-run",
                        default=None, action='store_true',
                        help="Only show what would be done "
                        "(default unless running under sudo")
    parser.add_argument("-f", "--force",
                        default=False, action='store_true',
                        help="Will move files even if destination exists")
    parser.add_argument("-c", "--clean",
                        default=False, action='store_true',
                        help="""Will remove other matches than the one that
                        gets promoted. In other words, useful when one name
                        has several matches and only the last one is desired.
                        Typically after a successful rsave""")
    parser.add_argument("image", type=str)
    args = parser.parse_args(argv)

    imagesrepo = ImagesRepo()
    return imagesrepo.share(
        args.image, args.alias, args.dry_run, args.force, args.clean)

####################


@subcommand
def leases(*argv):
    usage = """
    Unless otherwise specified, displays current leases, from today onwards
    """
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('-c', '--check', action='store_true', default=False,
                        help="Check if you currently have a lease")
    parser.add_argument('-i', '--interactive', action='store_true',
                        default=False,
                        help="Interactively prompt for commands "
                             "(create, update, delete)")
    args = parser.parse_args(argv)

    message_bus = asyncio.Queue()
    leases = Leases(message_bus)
    if args.check:
        access = check_reservation(leases, verbose=True)
        return 0 if access else 1
    else:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(leases.main(args.interactive))
        loop.close()
        return 0

####################


@subcommand
def monitornodes(*argv):                                # pylint: disable=r0914

    # xxx hacky - do a side effect in the logger module
    import rhubarbe.logger
    rhubarbe.logger.logger = rhubarbe.logger.monitor_logger
    from rhubarbe.logger import logger

    usage = """
    Cyclic probe all selected nodes, and reports
    real-time status at a sidecar service over websockets
    """
    config = Config()
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '-c', "--cycle",
        default=config.value('monitor', 'cycle_nodes'),
        type=float,
        help="Delay to wait between 2 probes of each node")
    parser.add_argument(
        "-u", "--sidecar-url", dest="sidecar_url",
        default=Config().value('sidecar', 'url'),
        help="url for the sidecar server")
    parser.add_argument(
        "-w", "--wlan", dest="report_wlan",
        default=False, action='store_true',
        help="ask for probing of wlan traffic rates")
    parser.add_argument("-v", "--verbose",
                        action='store_true', default=False)
    add_selector_arguments(parser)
    args = parser.parse_args(argv)

    selector = selected_selector(args)
    message_bus = asyncio.Queue()

    # xxx having to feed a Display instance with nodes
    # at creation time is a nuisance
    display = Display([], message_bus)

    logger.info({'selected_nodes': selector})
    monitornodes = MonitorNodes(selector.cmc_names(),
                                message_bus=message_bus,
                                cycle=args.cycle,
                                sidecar_url=args.sidecar_url,
                                report_wlan=args.report_wlan,
                                verbose=args.verbose)

    async def async_main():
        # run both the core and the log loop in parallel
        await asyncio.gather(monitornodes.run_forever(),
                             display.run())

    MonitorLoop("monitornodes").run(async_main(), logger)
    return 0

####################


@subcommand
def monitorphones(*argv):

    # xxx hacky - do a side effect in the logger module
    import rhubarbe.logger
    rhubarbe.logger.logger = rhubarbe.logger.monitor_logger
    from rhubarbe.logger import logger

    usage = """
    Cyclic probe all known phones, and reports real-time status
    at a sidecar service over websockets
    """
    config = Config()
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "-c", "--cycle",
        default=config.value('monitor', 'cycle_phones'),
        type=float,
        help="Delay to wait between 2 probes of each phone")
    parser.add_argument(
        "-u", "--sidecar-url", dest="sidecar_url",
        default=Config().value('sidecar', 'url'),
        help="url for the sidecar server")
    parser.add_argument(
        "-v", "--verbose", action='store_true')
    args = parser.parse_args(argv)

    logger.info("Using all phones")
    monitorphones = MonitorPhones(**vars(args))

    MonitorLoop("monitorphones").run(
        monitorphones.run_forever(),
        logger)
    return 0

####################


@subcommand
def monitorleases(*argv):

    # xxx hacky - do a side effect in the logger module
    import rhubarbe.logger
    rhubarbe.logger.logger = rhubarbe.logger.monitor_logger
    from rhubarbe.logger import logger

    usage="""
    Cyclic check of leases; also reacts to 'request' messages on
    the sidecar channel, which triggers leases acquisition right away.
    See config for defaults.
    """
    config = Config()
    parser = ArgumentParser(
        usage=usage, formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "-u", "--sidecar-url", dest="sidecar_url",
        default=Config().value('sidecar', 'url'),
        help="url for the sidecar server")
    parser.add_argument(
        "-v", "--verbose", default=False, action='store_true')
    args = parser.parse_args(argv)

    message_bus = asyncio.Queue()


    monitorleases = MonitorLeases(
        message_bus, args.sidecar_url, args.verbose)

    MonitorLoop("monitorleases").run(
        monitorleases.run_forever(),
        logger)
    return 0

####################


@subcommand
def accountsmanager(*argv):

    usage = "The core of the accounts manager; reserved to root"

    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-c", "--cycle",
                        help="Set cycle in seconds; 0 means run only once;"
                             " default from config file.",
                        default=None)
    args = parser.parse_args(argv)

    accounts_manager = AccountsManager()
    return accounts_manager.main(args.cycle)

####################


@subcommand
def inventory(*argv):
    usage = """
    Display inventory
    """
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.parse_args(argv)
    inventory = Inventory()
    inventory.display(verbose=True)
    return 0

####################


@subcommand
def config(*argv):
    usage = """
    Display global configuration
    """
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("sections", nargs='*',
                        type=str,
                        help="config section(s) to display")
    args = parser.parse_args(argv)
    config = Config()
    config.display(args.sections)
    return 0

####################


@subcommand
def template(*argv):
    usage = """
    Show the template file to create the nodes and phones inventory
    under /etc/rhubarbe
    """
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "-n", "--nodes", dest='nodes',
        action='store_true', default=False,
        help="Show template for /etc/rhubarbe/inventory-nodes.json")
    parser.add_argument(
        "-p", "--phones", dest='phones',
        action='store_true', default=False,
        help="Show template for /etc/rhubarbe/inventory-phones.json")
    args = parser.parse_args(argv)

    def show_template(nodes_or_phones):
        template = resource_string(
            'rhubarbe', f"config/inventory-{nodes_or_phones}.json.template")
        print("# =========="
              f" template for /etc/rhubarbe/inventory-{nodes_or_phones}.json")
        print(template.decode(encoding='utf-8'))

    # pick -n by default
    if not args.nodes and not args.phones:
        args.nodes = True

    for nodes_or_phones in ('nodes', 'phones'):
        selected = getattr(args, nodes_or_phones)
        if selected:
            show_template(nodes_or_phones)

####################


@subcommand
def version(*_):
    from rhubarbe.version import __version__
    print(f"rhubarbe version {__version__}")
    return 0



"""
The Node object is the handle for doing most CMC-oriented actions like
querying status, turning on or off, and similar
"""

# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111, w0703, w1202

import os.path

import asyncio
import aiohttp

from rhubarbe.logger import logger
from rhubarbe.config import Config
from rhubarbe.inventory import Inventory
from rhubarbe.frisbee import Frisbee
from rhubarbe.imagezip import ImageZip


class Node:                                             # pylint: disable=r0902

    """
    This class allows to talk to various parts of a node
    created from the cmc hostname for convenience
    the inventory lets us spot the other parts (control, essentially)
    """
    def __init__(self, cmc_name, message_bus):
        self.cmc_name = cmc_name
        self.message_bus = message_bus
        self.status = None
        self.action = None
        self.mac = None
        # for monitornodes
        self.id = int("".join([x for x in cmc_name      # pylint: disable=c0103
                               if x in "0123456789"]))

    def __repr__(self):
        return f"<Node {self.control_hostname()}>"

    def is_known(self):
        return self.control_mac_address() is not None

    def control_mac_address(self):
        the_inventory = Inventory()
        return the_inventory.attached_hostname_info(self.cmc_name,
                                                    'control', 'mac')

    def control_ip_address(self):
        the_inventory = Inventory()
        return the_inventory.attached_hostname_info(self.cmc_name,
                                                    'control', 'ip')

    def control_hostname(self):
        the_inventory = Inventory()
        return the_inventory.attached_hostname_info(self.cmc_name,
                                                    'control', 'hostname')

    async def get_status(self):
        """
        returns self.status
        either 'on' or 'off', or None if something wrong is going on
        """
        result = await self._get_cmc_verb('status')
        return result

    async def turn_on(self):
        """
        turn node on; expected result would be 'ok' if it goes fine
        """
        result = await self._get_cmc_verb('on')
        return result

    async def turn_off(self):
        """
        turn node on; expected result would be 'ok' if it goes fine
        """
        result = await self._get_cmc_verb('off')
        return result

    async def do_reset(self):
        """
        turn node on; expected result would be 'ok' if it goes fine
        """
        result = await self._get_cmc_verb('reset')
        return result

    async def get_info(self):
        """
        turn node on; expected result would be 'ok' if it goes fine
        """
        result = await self._get_cmc_verb('info', strip_result=False)
        return result

    async def get_usrpstatus(self):
        """
        returns self.usrpstatus
        either 'on' or 'off', or None if something wrong is going on
        """
        result = await self._get_cmc_verb('usrpstatus')
        return result

    async def turn_usrpon(self):
        """
        turn on node's USRP; expected result would be 'ok' if it goes fine
        """
        result = await self._get_cmc_verb('usrpon')
        return result

    async def turn_usrpoff(self):
        """
        turn off node's USRP; expected result would be 'ok' if it goes fine
        """
        result = await self._get_cmc_verb('usrpoff')
        return result

    async def _get_cmc_verb(self, verb, strip_result=True):
        """
        verb typically is 'status', 'on', 'off' or 'info'
        """
        url = f"http://{self.cmc_name}/{verb}"
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as response:
                    text = await response.text(encoding='utf-8')
                    if strip_result:
                        text = text.strip()
                    setattr(self, verb, text)
        except aiohttp.client_exceptions.ClientConnectorError:
            logger.info(f"cannot connect to {url}")
            setattr(self, verb, None)
            return None
        except Exception:
            import traceback
            traceback.print_exc()
            setattr(self, verb, None)
            return None
        return getattr(self, verb)

    ####################
    # what status to expect after a message is sent
    expected_map = {
        'on': 'on',
        'reset': 'on',
        'off': 'off'
    }

    async def send_action(self, message="on", check=False, check_delay=1.):
        """
        Actually send action message like 'on', 'off' or 'reset'

        If check is True, waits for check_delay seconds
        before checking again that the status is what is expected, i.e.
        | message  | expected |
        |----------|----------|
        | on,reset | on       |
        | off      | off      |

        return value stored in self.action:

        * if check is false
          * True if request can be sent and returns 'ok',
            or None if something goes wrong
        * otherwise:
          * True to indicate that the node is correctly
            in 'on' mode after checking
          * False to indicate that the node is 'off' after checking
          * None if something goes wrong
        """
        url = f"http://{self.cmc_name}/{message}"
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as response:
                    text = await response.text(encoding='utf-8')
        except Exception:
            self.action = None
            return self

        is_ok = text.strip() == 'ok'

        if not check:
            self.action = is_ok
            return self
        await asyncio.sleep(check_delay)
        await self.get_status()
        self.action = self.status == self.expected_map[message]
        return self

    ####################
    message_to_reset_map = {'on': 'reset', 'off': 'on'}

    async def feedback(self, field, message):
        await self.message_bus.put(
            {'ip': self.control_ip_address(), field: message})

    async def ensure_reset(self):
        if self.status is None:
            await self.get_status()
        if self.status not in self.message_to_reset_map:
            await self.feedback(
                'reboot', f"Cannot get status at {self.cmc_name}")
        message_to_send = self.message_to_reset_map[self.status]
        await self.feedback(
            'reboot', f"Sending message '{message_to_send}' to CMC {self.cmc_name}")
        await self.send_action(message_to_send, check=True)
        if not self.action:
            await self.feedback(
                'reboot', f"Failed to send message {message_to_send} to CMC {self.cmc_name}")

    # used to be a coroutine, but since we need this
    # when dealing by KeybordInterrupt, it appears much safer
    # to just keep it a traditional function
    def manage_nextboot_symlink(self, action):
        """
        Messes with the symlink in /tftpboot/pxelinux.cfg/
        Depending on 'action'
        * 'cleanup' or 'harddrive' : clear the symlink
          corresponding to this CMC
        * 'frisbee' : define a symlink so that next boot
          will run the frisbee image
        see rhubarbe.conf for configurable options
        """

        the_config = Config()
        root = the_config.value('pxelinux', 'config_dir')
        frisbee = the_config.value('pxelinux', 'frisbee_image')

        # of the form 01-00-03-1d-0e-03-53
        mylink = "01-" + self.control_mac_address().replace(':', '-')
        source = os.path.join(root, mylink)

        if action in ('cleanup', 'harddrive'):
            if os.path.exists(source):
                logger.info(f"Removing {source}")
                os.remove(source)
        elif action in ('frisbee', ):
            if os.path.exists(source):
                os.remove(source)
            logger.info(f"Creating {source}")
            os.symlink(frisbee, source)
        else:
            logger.critical(
                f"manage_nextboot_symlink : unknown action {action}")


    ##########
    async def wait_for_telnet(self, service):
        ipaddr = self.control_ip_address()
        if service == 'frisbee':
            self.frisbee = Frisbee(ipaddr, self.message_bus)
            await self.frisbee.wait_until_connect()
        elif service == 'imagezip':
            self.imagezip = ImageZip(ipaddr, self.message_bus)
            await self.imagezip.wait_until_connect()

    async def reboot_on_frisbee(self, idle):
        self.manage_nextboot_symlink('frisbee')
        await self.ensure_reset()
        await self.feedback('reboot', f"idling for {idle}s")
        await asyncio.sleep(idle)

    async def run_frisbee(self, ipaddr, port, reset):
        await self.wait_for_telnet('frisbee')
        self.manage_nextboot_symlink('cleanup')
        result = await self.frisbee.run(ipaddr, port)
        #logger.info(f"run_frisbee -> {result}")
        if reset:
            await self.ensure_reset()
        else:
            await self.feedback('reboot',
                                'skipping final reset')
        return result

    async def run_imagezip(self, port, reset, radical, comment):
        await self.wait_for_telnet('imagezip')
        self.manage_nextboot_symlink('cleanup')
        result = await self.imagezip.run(port, self.control_hostname(),
                                         radical, comment)
        #logger.info(f"run_imagezip -> {result}")
        if reset:
            await self.ensure_reset()
        else:
            await self.feedback('reboot',
                                'skipping final reset')
        return result



#!/usr/bin/env python3

"""
determine local IP address to use for frisbeed
"""

# c0111 no docstrings yet
# w0201 attributes defined outside of __init__
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111

import subprocess
import re
import ipaddress

MATCHER = re.compile(r"inet (?P<address>([0-9]+\.){3}[0-9]+)/(?P<mask>[0-9]+)")

_LOCAL_INTERFACES = None


def local_interfaces():
    global _LOCAL_INTERFACES                            # pylint: disable=w0603
    if _LOCAL_INTERFACES is not None:
        return _LOCAL_INTERFACES
    ip_links = subprocess.Popen(
        ['ip', 'address', 'show'],
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
        universal_newlines=True
    )
    _LOCAL_INTERFACES = []
    out, _ = ip_links.communicate()
    for line in out.split("\n"):
        line = line.strip()
        match = MATCHER.match(line)
        if match:
            interface = ipaddress.ip_interface(
                f"{match.group('address')}/{match.group('mask')}")
            if not interface.is_loopback:
                _LOCAL_INTERFACES.append(interface)
    return _LOCAL_INTERFACES


def local_ip_on_same_network_as(peer):
    """
    Typically if peer is 192.168.3.1 and we have an interface 192.168.3.200/24
    then this will return a tuple of strings 192.168.3.200, 24
    """
    for interface in local_interfaces():
        length = interface.network.prefixlen
        peer_interface = ipaddress.ip_interface(f"{peer}/{length}")
        if peer_interface.network == interface.network:
            return str(interface.ip), str(length)
    return None


if __name__ == '__main__':
    LOCAL_IP, MASK = local_ip_on_same_network_as("192.168.3.1")
    print(f"found {LOCAL_IP}/{MASK}")



"""
This implements a daemon that runs on faraday and that
takes care of the defined / authorized logins

think of it as a dedicated nodemanager
"""

# XXX:
# current implementation drawbacks:
# when migrating from an old box to a new box, it would make sense
# for this tool to enforce things that are metastable
# that is to say, moving stuff over from one box to the other
# can result in the homedir being present but without the .ssh dir, or
# with wrong ownership..
# This currently is not well taken care of, because all is done once when
# creating the account


# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111,w0703,w1202,w1202

import time
import os
import pwd

from pathlib import Path

from rhubarbe.logger import accounts_logger as logger
from rhubarbe.config import Config
from rhubarbe.plcapiproxy import PlcApiProxy


####################
# adapted from historical planetlab nodemanager

def replace_file_with_string(destination_path, new_contents,
                             owner=None, chmod=None,
                             remove_if_empty=False):
    """
    Replace a file with new contents
    checks for changes
      does not do anything if previous state was already right
    can handle chmod/chown if requested
    can also remove resulting file if contents are void, if requested

    returns
      * True if a change occurred, or the file is deleted
      * False if the file has no change
      * None if the file could not be created (e.g. directory missing)
    """
    try:
        with destination_path.open() as previous:
            current = previous.read()
    except IOError:
        current = ""
    if current == new_contents:
        # if turns out to be an empty string, and remove_if_empty is set,
        # then make sure to trash the file if it exists
        if remove_if_empty and not new_contents and destination_path.is_file():
            logger.info(
                "replace_file_with_string: removing file {}"
                .format(destination_path))
            try:
                destination_path.unlink()
            finally:
                return True                             # pylint: disable=w0150
        # we're done and have nothing to do
        return False
    # overwrite file: create a temp in the same directory
    try:
        with destination_path.open('w') as new:
            new.write(new_contents)
        if chmod:
            destination_path.chmod(chmod)
        if owner:
            os.system("chown {} {}".format(owner, destination_path))
        return True
    except IOError as exc:
        logger.error("Cannot create {}".format(destination_path))
        return None

####################


class AccountsManager:

    def __init__(self):
        the_config = Config()
        self.plcapiurl = the_config.value('plcapi', 'url')
        self.email = the_config.value('plcapi', 'admin_email')
        self.password = the_config.value('plcapi', 'admin_password')

        self._proxy = None

    # not reconnectable for now
    def proxy(self):
        if self._proxy is None:
            # also set debug=True if needed
            self._proxy = PlcApiProxy(self.plcapiurl,
                                      email=self.email,
                                      password=self.password)
        return self._proxy

    @staticmethod
    def slices_from_passwd():
        """
        Inspect /etc/passwd and return all logins

        ignore the ones that do not have a '_' in them so that
        we leave alone individual accounts
        """
        return [record.pw_name for record in pwd.getpwall()
                if '_' in record.pw_name
                and Path(record.pw_dir).exists()]

    @staticmethod
    def slices_with_authorized():
        """
        Iterator:

        Inspect /home/ to find accounts that have authorized_keys.

        Focus on the ones that have a '_' in them, so that we leave
        alone custom accounts like 'guest' or similar.
        """
        homeroot = Path('/home')
        for authorized in homeroot.glob('*/.ssh/authorized_keys'):
            # need to move 2 steps up
            basename = authorized.parts[-3]
            if '_' not in basename:
                continue
            yield basename

    @staticmethod
    def create_account(slicename):
        """
        Does useradd with the right options
        Plus, creates an empty .ssh dir with proper permissions
        NOTE that this addresses ubuntu for now, fedora 'useradd'
        being slightly different as far as I remember
        (at least wrt homedir creation, IIRC again)
        """
        commands = [
            "useradd --create-home --user-group {x} --shell /bin/bash",
            "mkdir /home/{x}/.ssh",
            "chmod 700 /home/{x}/.ssh",
            "chown -R {x}:{x} /home/{x}",
        ]
        for cmd in commands:
            command = cmd.format(x=slicename)
            logger.info("Running {}".format(command))
            retcod = os.system(command)
            if retcod != 0:
                logger.error("{} -> {}".format(command, retcod))

    @staticmethod
    def create_ssh_config(slicename):
        """
        Initialize slice's .ssh/config that keeps ssh from
        being too picky with host keys and similar

        Performed only if not yet existing
        """
        ssh_config_file = Path("/home") / slicename / ".ssh/config"

        # define the magic sequence for both fit* and data*
        config_bases = ['fit', 'data']
        #
        config_pattern = """Host {base}*
StrictHostKeyChecking no
UserKnownHostsFile=/dev/null
CheckHostIP=no
"""

        ssh_config = "\n".join(
            [config_pattern.format(base=base) for base in config_bases])
        replace_file_with_string(ssh_config_file,
                                 ssh_config,
                                 chmod=0o600,
                                 owner="{x}:{x}".format(x=slicename))

    @staticmethod
    def apply_keys(slicename, keys_string):
        auth_path = Path("/home") / slicename / ".ssh/authorized_keys"
        replace_file_with_string(auth_path,
                                 keys_string,
                                 chmod=0o600,
                                 owner="{x}:{x}".format(x=slicename),
                                 remove_if_empty=True)

    ##########
    @staticmethod
    def authorized_key_lines(plc_slice, plc_persons_by_id, plc_keys_by_id):
        """
        returns the expected contents of that slice's authorized_keys file
        """
        persons = [plc_persons_by_id[id] for id in plc_slice['person_ids']]
        key_ids = set(sum((p['key_ids'] for p in persons), []))
        # somtimes the key comes with a final "\n"
        keys = [plc_keys_by_id[id] for id in key_ids]
        key_lines = [(k['key'].replace("\n", "") + "\n") for k in keys]
        # this is so we get something canonical that does not
        # change everytime
        key_lines.sort()
        return "".join(key_lines)

    ##########
    def manage_accounts(self, policy):             # pylint: disable=r0914

        # get plcapi specification of what should be
        slices = self.proxy().GetSlices(
            {}, ['slice_id', 'name', 'expires', 'person_ids'])
        persons = self.proxy().GetPersons(
            {}, ['person_id', 'email', 'slice_ids', 'key_ids'])
        keys = self.proxy().GetKeys()

        current_leases = []
        if policy == 'leased':
            now = int(time.time())
            current_leases = self.proxy().GetLeases(
                {'alive': now}, ['name'])

        if (current_leases is None or slices is None
                or persons is None or keys is None):
            logger.info("PLCAPI unreachable - back to sleep")
            return

        # prepare data
        persons_by_id = {p['person_id']: p for p in persons}
        keys_by_id = {k['key_id']: k for k in keys}
        # current_slicenames will contain 0 or 1 item
        current_slicenames = [lease['name'] for lease in current_leases]

        # initialize with the slice names that are in /etc/passwd
        logins = self.slices_from_passwd()

        # initialize map login_name -> authorized_keys contents
        # this is where we handle the fact that obsolete slices
        # will effectively have their authorized_keys voided
        auths_by_login = {login: "" for login in logins}

        for sliceobj in slices:
            slicename = sliceobj['name']
            # policy-dependant
            if policy == 'closed':
                authorized_keys = ""

            elif policy == 'leased':
                authorized_keys = ""
                if slicename in current_slicenames:
                    authorized_keys = self.authorized_key_lines(
                        sliceobj, persons_by_id, keys_by_id)

            # policy == 'open'
            else:
                authorized_keys = self.authorized_key_lines(
                    sliceobj, persons_by_id, keys_by_id)

            auths_by_login[slicename] = authorized_keys

        # implement it
        for slicename, keys in auths_by_login.items():
            try:
                if slicename not in logins:
                    self.create_account(slicename)
                # do this always, allows to propagate later changes
                self.create_ssh_config(slicename)
                self.apply_keys(slicename, keys)
            except Exception:
                logger.exception("Could not deal with slice {}"
                                 .format(slicename))

    def run_forever(self, cycle, policy):
        while True:
            beg = time.time()
            logger.info("---------- rhubarbe accounts manager "
                        "policy = {}, cycle {}s"
                        .format(policy, cycle))
            self.manage_accounts(policy)
            now = time.time()
            duration = now - beg
            towait = cycle - duration
            if towait > 0:
                logger.info("---------- rhubarbe accounts manager - "
                            "sleeping for {:.2f}s"
                            .format(towait))
                time.sleep(towait)
            else:
                logger.info("duration {}s exceeded cycle {}s - "
                            "skipping sleep"
                            .format(duration, cycle))

    def main(self, cycle):
        """
        cycle is the duration in seconds of one cycle

        Corner cases:
        * cycle = None : fetch value from config_bases
        * cycle = 0 : run just once (for debug mostly)
        """

        if cycle is None:
            cycle = Config().value('accounts', 'cycle')
        cycle = int(cycle)
        policy = Config().value('accounts', 'access_policy')
        if policy not in ('open', 'leased', 'closed'):
            logger.error("Unknown policy {} - using 'closed'"
                         .format(policy))
            policy = 'closed'
        # trick is
        if cycle != 0:
            self.run_forever(cycle, policy)
        else:
            logger.info("---------- rhubarbe accounts manager oneshot "
                        "policy = {}"
                        .format(policy))
            self.manage_accounts(policy)



"""
create/configure logger objects
"""

# c0111 no docstrings yet
# c0103 constants names should be UPPER_CASE
# w0201 attributes defined outside of __init__
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# r0903 too few public methods
# pylint: disable=c0103, w0703

import sys
from pathlib import Path

import logging
import logging.config

# with systemd it's sooo simpler to log on stdout, that gets managed by journal
# so, we essentially need
# * one all-purpose logger that goes into $HOME/rhubarbe.log
# * one special logger for monitor*s that goes onto stdout -> journal
# * one special logger for accounts - ditto but with a shorter layout

rhubarbe_logging_config = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'standard': {
            'format': '%(asctime)s %(levelname)s '
                      '%(filename)s:%(lineno)d %(message)s',
            'datefmt': '%m-%d %H:%M:%S'
        },
        'shorter': {
            'format': '%(asctime)s %(levelname)s %(message)s',
            'datefmt': '%d %H:%M:%S'
        },
    },
    'handlers': {
        'rhubarbe': {
            'level': 'INFO',
            'class': 'logging.FileHandler',
            'formatter': 'standard',
            'filename': 'rhubarbe.log',
        },
        'monitor': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'standard',
            'stream': sys.stdout,
        },
        'accounts': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'shorter',
            'stream': sys.stdout,
        },
    },
    'loggers': {
        'monitor': {
            'handlers': ['monitor'],
            'level': 'INFO',
            'propagate': False,
        },
        'accounts': {
            'handlers': ['accounts'],
            'level': 'INFO',
            'propagate': False,
        },
        'rhubarbe': {
            'handlers': ['rhubarbe'],
            'level': 'INFO',
            'propagate': False,
        },
    },
}

logging.config.dictConfig(rhubarbe_logging_config)

# general case:
# from rhubarbe.logger import logger
logger = logging.getLogger('rhubarbe')

# monitor
# from rhubarbe.logger import monitor_logger as logger
monitor_logger = logging.getLogger('monitor')

# accounts
# from rhubarbe.logger import accounts_logger as logger
accounts_logger = logging.getLogger('accounts')

####################
# test
if __name__ == '__main__':
    logger.info("in rhubarbe")
    monitor_logger.info("in monitor")



# pylint: disable=c0111, w1202

import time
import asyncio

from rhubarbe.logger import monitor_logger as logger
from rhubarbe.config import Config
from rhubarbe.leases import Leases

from rhubarbe.monitor.reconnectable import ReconnectableSidecar

class MonitorLeases:                                    # pylint: disable=r0902

    def __init__(self, message_bus, sidecar_url,        # pylint: disable=r0913
                 verbose=False):
        self.message_bus = message_bus
        self.sidecar_url = sidecar_url
        self.verbose = verbose

        # websockets
        self.reconnectable = \
            ReconnectableSidecar(sidecar_url, 'leases')

        self.cycle = float(Config().value('monitor', 'cycle_leases'))
        self.step = float(Config().value('monitor', 'step_leases'))


    def on_back_channel(self, umbrella):
        # when anything is received on the backchannel, we go to fast track
        logger.info(f"MonitorLeases.on_back_channel, umbrella={umbrella}")
        self.fast_track = True                          # pylint: disable=w0201


    async def mainloop(self):
        leases = Leases(self.message_bus)
        if self.verbose:
            logger.info("Entering monitor on leases")
        while True:
            self.fast_track = False                     # pylint: disable=w0201
            trigger = time.time() + self.cycle
            # check for back_channel every 50 ms
            while not self.fast_track and time.time() < trigger:
                await asyncio.sleep(self.step)

            try:
                if self.verbose:
                    logger.info("monitorleases mainloop")
                await leases.refresh()
                # xxx this is fragile
                omf_leases = leases.resources
                logger.info("advertising {} leases".format(len(omf_leases)))
                await self.reconnectable.emit_infos(omf_leases)
                if self.verbose:
                    logger.info("Leases details: {}".format(omf_leases))
            except Exception:
                logger.exception("monitornodes could not get leases")

    async def run_forever(self):
        def closure(umbrella):
            return self.on_back_channel(umbrella)
        await asyncio.gather(
            self.mainloop(),
            self.reconnectable.keep_connected(),
            self.reconnectable.watch_back_channel('leases', closure)
        )



#!/usr/bin/env python3

import sys
import os

# use line buffering on stdout and stderr
sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 1) # no buffering
sys.stderr = os.fdopen(sys.stderr.fileno(), 'w', 1) # no buffering

import rhubarbe.main
from rhubarbe.selector import MisformedRange

def main():
    supported = rhubarbe.main.supported_subcommands
    command = sys.argv[0]
    args = None
    subcommand = None
    # can we find a supported subcommand in argv[0]
    # i.e. when invoked as rhubarbe-monitornodes
    for candidate in supported:
        # just checking 'candidate in sys.argv[0]'
        # does not work as it would e.g. find subcommand
        # 'on' when trying to call monitornodes...
        # instead, we check that the command ends with e.g. -monitornodes
        if sys.argv[0].endswith("-" + candidate):
            subcommand = candidate
            args = sys.argv[1:]
            break

    # not found that way : subcommand must be argv[1]
    # so we need at least one argv left
    if not subcommand and len(sys.argv) <= 1:
        print(f"{command} needs a subcommand in {{{','.join(supported)}}}")
        exit(1)

    if not subcommand:
        subcommand = sys.argv[1]
        args = sys.argv[2:]
        if subcommand not in supported:
            print(f"Unknown subcommand {subcommand} "
                  f"- use one among {{{','.join(supported)}}}")
            exit(1)
    # do it
    entry_point = getattr(rhubarbe.main, subcommand)
    # remove subcommand from args
    try:
        exit(entry_point(*args))
    except MisformedRange as e:
        print("ERROR: ", e)
        exit(1)
    except Exception as exc:
        import traceback
        traceback.print_exc()
        print("{command} {subcommand} : Something went badly wrong : {exc}")
        exit(1)

if __name__ == '__main__':
    main()



"""
Parsing the output of a remote imagezip process
for animating rhubarbe save
"""

# c0111 no docstrings yet
# w0201 attributes defined outside of __init__
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111, c0103, w1201, w1202, w1203

import os
import time
import asyncio

from rhubarbe.logger import logger
from rhubarbe.config import Config
from rhubarbe.telnet import TelnetProxy


class ImageZip(TelnetProxy):

    async def ticker(self):
        await asyncio.sleep(0.5)
        while self.running:
            await self.feedback('tick', '')
            await asyncio.sleep(0.15)
        await self.feedback('tick', 'END')

    # we don't parse anything here because there does not seem to be a way
    # to estimate some total first off, and later get percentages
    # useful to see the logs:
    # self.feedback('imagezip_raw', line)
    # send 10 ticks in a raw - not a good idea
    # for i in range(10): self.feedback('tick', '')
    def line_callback(self, line):
        pass


    async def run(self, port, nodename,                 # pylint: disable=r0914
                  radical, comment):
        the_config = Config()
        server_ip = the_config.local_control_ip()
        imagezip = the_config.value('frisbee', 'imagezip')
        netcat = the_config.value('frisbee', 'netcat')
        # typically /dev/sda
        hdd = the_config.value('frisbee', 'hard_drive')
        # typically /dev/sda1
        root_partition = the_config.value('frisbee', 'root_partition')
        commands = []
        if root_partition and root_partition.lower() != 'none':
            command = ""
            # Managing the /etc/rhubarbe-image stamp
            # typically /mnt
            mount_point = the_config.value('frisbee', 'mount_point')
            date = time.strftime("%Y-%m-%d@%H:%M", time.localtime())
            who = os.getlogin()
            # create mount point if needed
            command += f'[ -d {mount_point} ] || mkdir {mount_point}; '
            # mount it, and only if successful ...
            command += f'mount {root_partition} {mount_point} && '
            # add to the stamp, and umount
            # beware of {{ and }} as these are formats
            command += (f'{{ echo "{date} - node {nodename} '
                        f' - image {radical} - by {who}"')
            if comment:
                command += f'" - {comment}"'
            command += f' >> {mount_point}/etc/rhubarbe-image ; '
            command += f'umount {mount_point}; }} ; '
            commands.append(command)

        # need to set pipefail so that we capture an error if ANY
        # of the two commands in the pipe fail
        commands.append("set -o pipefail")
        command = f"{imagezip} -o -z1 {hdd} - | {netcat} {server_ip} {port}"
        commands.append(command)
        logger.info(f"on {self.control_ip} : running command {command}")

        await self.feedback('frisbee_status',
                            f"starting imagezip on {self.control_ip}")

        # print out exit status so the parser can catch it and expose it
        retcod, _ = await asyncio.gather(
            self.session(commands),
            self.ticker(),
        )
        logger.info(f"imagezip on {self.control_ip} returned {retcod}")

        return retcod



"""
The collector is used when saving an images
it starts a local netcat process that binds a specific port
and stores everything on the newly saved image file
"""

import asyncio

from rhubarbe.logger import logger
from rhubarbe.config import Config

# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111,w1202,r1705


class Collector:
    def __init__(self, image, message_bus):
        self.image = image
        self.message_bus = message_bus
        #
        self.subprocess = None
        self.port = None

    async def feedback(self, field, msg):
        await self.message_bus.put({field: msg})

    def feedback_nowait(self, field, msg):
        self.message_bus.put_nowait({field: msg})

    async def start(self):                              # pylint: disable=r0914
        """
        Start a collector instance; returns a port_number
        """
        the_config = Config()
        netcat = the_config.value('frisbee', 'netcat')
        local_ip = the_config.local_control_ip()

        # should use default.ndz if not provided
        # use shell-style as we rather have bash handle the redirection
        # we instruct bash to exec nc;
        # otherwise when cleaning up we just kill the bash process
        # but nc is left lingering behind
        # WARNING: it is intended that format contains {port}
        # for future formatting
        command_format_ubuntu = (
            f"exec {netcat} -d -l {local_ip} {{port}} > {self.image}"
            f" 2> /dev/null")
        command_format_fedora = (
            f"exec {netcat}    -l {local_ip} {{port}} > {self.image}"
            f" 2> /dev/null")

        netcat_style = the_config.value('frisbee', 'netcat_style')
        if netcat_style not in ('fedora', 'ubuntu'):
            message = f"wrong netcat_style {netcat_style}"
            print(message)
            raise Exception(message)
        command_format = (
            command_format_fedora if netcat_style == 'fedora'
            else command_format_ubuntu)

        nb_attempts = int(the_config.value('networking', 'pattern_size'))
        pat_port = the_config.value('networking', 'pattern_port')
        for i in range(1, nb_attempts+1):
            pat = str(i)
            port = str(eval(                            # pylint: disable=w0123
                pat_port.replace('*', pat)))
            command = command_format.format(port=port)
            self.subprocess = await asyncio.create_subprocess_shell(command)
            await asyncio.sleep(1)
            # after such a short time, frisbeed should not have returned yet
            # if is has, we try our luck on another couple (ip, port)
            command_line = command
            if self.subprocess.returncode is None:
                logger.info(f"collector started: {command_line}")
                await self.feedback(
                    'info', f"collector started on {self.image}")
                self.port = port
                return port
            else:
                logger.warning(
                    f"failed to start collector with {command_line}")
        logger.critical("Could not find a free port to start collector")
        raise Exception("Could not start collector server")

#    async def stop(self):
#        self.stop_nowait()

    def stop_nowait(self):
        # make it idempotent
        if self.subprocess:
            # when everything is running fine, nc will exit on its own
            try:
                self.subprocess.kill()
            except Exception:                           # pylint: disable=w0703
                pass
            self.subprocess = None
            logger.info(f"collector (on port {self.port}) stopped")
            self.feedback_nowait(
                'info', f"image collector server (on port {self.port}) stopped")



# pylint: disable=c0111

import websockets
import json
import asyncio

from r2lab import SidecarAsyncClient

from rhubarbe.logger import monitor_logger as logger

#import logging
#logger.setLevel(logging.DEBUG)

class ReconnectableSidecar:

    def __init__(self, url, category, keep_period=1):
        # keep_period is the frequency where connection is verified for open-ness
        self.url = url
        self.category = category
        self.keep_period = keep_period
        # caller MUST run keep_connected() 
        self.proto = None
        self.counter = 0
        logger.info(f"reconnectable sidecar to {url} ")

    
    async def emit_info(self, info):
        # create a list with that one info
        return await self.emit_infos([info])


    async def emit_infos(self, infos):
        if not self.proto:
            logger.warning(f"dropping message {infos}")
            return False
        logger.debug(f"Sending {infos}")
        # xxx use Payload
        payload = dict(category=self.category, action='info', message=infos)
        # xxx try/except here
        try:
            await self.proto.send(json.dumps(payload))
            self.counter += 1
        except ConnectionRefusedError:
            logger.warning(f"Could not send {self.category} - dropped")
        except Exception as exc:
            # xxx to review
            logger.exception("send failed")
            self.proto = None
            return False
        return True


    async def keep_connected(self):
        """
        A continuous loop that keeps the connection open
        """
        while True:
            logger.debug(f"in keep_connected, proto={self.proto}")
            if self.proto and self.proto.open:
                pass
            else:
                # xxx should we close() our client ?
                self.proto = None
                # see if we need ssl
                secure, *_ = websockets.uri.parse_uri(self.url)
                kwds = {}
                if secure:
                    import ssl
                    kwds.update(dict(ssl=ssl.SSLContext()))
                try:
                    logger.info(f"(re)-connecting to {self.url} ...")
                    self.proto = await SidecarAsyncClient(self.url, **kwds)
                    logger.debug("connected !")
                except ConnectionRefusedError:
                    logger.warning(f"Could not connect to {self.url} at this time")
                except:
                    logger.exception(f"Could not connect to {self.url} at this time")
            await asyncio.sleep(self.keep_period)
                  

    async def watch_back_channel(self, category, callback):
        while True:
            if not self.proto:
                logger.debug(f"backing off for {self.keep_period}s")
                await asyncio.sleep(self.keep_period)
                continue
            try:
                incoming = await self.proto.recv()
                umbrella = json.loads(incoming)
                logger.info(f"tmp - got incoming {umbrella['category']} x {umbrella['action']}")
                if (umbrella['category'] == category and 
                        umbrella['action'] == 'request'):
                    callback(umbrella)
            except Exception as exc:
                ### to review
                logger.exception("recv failed .. fix me")
                self.proto = None



#!/usr/bin/env python3

"""
packaging / installing
"""

# c0103: should use uppercase names
# c0326: no space allowed around keyword argument assignment
# pylint: disable=c0103,c0326

import setuptools

from rhubarbe.version import __version__

LONG_DESCRIPTION = \
    "See https://github.com/parmentelat/rhubarbe/blob/master/README.md"

# subcommands
#
# avoid this; we may need this early, when requirements are not met yet
# from rhubarbe.main import supported_subcommands
# instead, cut'n'paste from the rhubarbe help message
rhubarbe_help = (
    "nodes,status,on,off,reset,info,usrpstatus,usrpon,usrpoff,"
    "load,save,wait,images,resolve,share,"
    "inventory,config,template,version,"
    "monitornodes,monitorphones,monitorleases,accountsmanager"
)
supported_subcommands = rhubarbe_help.split(",")

# requirements
#
# *NOTE* for ubuntu: also run this beforehand
# apt-get -y install libffi-dev
# which is required before pip can install asyncssh
INSTALL_REQUIRES = [
    'telnetlib3',
    'aiohttp',
    'asyncssh',
    'progressbar33',
    # for monitors
    'asynciojobs',
    'r2lab',
    # not yet used
    'aioxmlrpc',
]

# add convenience entry points like rhubarbe-load
all_commands = (
    ['rhubarbe'] +
    ['rhubarbe-{}'.format(subcommand)
     for subcommand in supported_subcommands])

setuptools.setup(
    name="rhubarbe",
    author="Thierry Parmentelat",
    author_email="thierry.parmentelat@inria.fr",
    description="Testbed Management Framework for R2Lab",
    long_description=LONG_DESCRIPTION,
    license="CC BY-SA 4.0",
    keywords=['R2lab', 'networking testbed'],

    packages=['rhubarbe', 'rhubarbe.monitor'],
    version=__version__,
    python_requires=">=3.5",

    entry_points={
        'console_scripts': [
            '{} = rhubarbe.__main__:main'
            .format(command) for command in all_commands
        ]
    },
    package_data={
        'rhubarbe': ['config/*.conf', 'config/*.template'],
    },

    install_requires=INSTALL_REQUIRES,

    project_urls={
        'source': "https://github.com/parmentelat/rhubarbe/",
    },

    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Information Technology",
        "Programming Language :: Python :: 3.5",
    ],
)
